#include <M5Core2.h>
#include <Wire.h>
#define I2C_ADDR_AXP192         0x34
#define I2C_ADDR_FT6336         0x38

#define ADDR_DEVICE_MODE        0x00
#define ADDR_TD_STATUS          0x02

#define ADDR_TOUCH_ID1          0x05
#define ADDR_TOUCH_X1           0x03
#define ADDR_TOUCH_Y1           0x05

#define ADDR_TOUCH_ID2          0x0B
#define ADDR_TOUCH_X2           0x09
#define ADDR_TOUCH_Y2           0x0B

// ---------------------------------
#define ADDR_THRESHOLD          0x80
#define ADDR_FILTER_COE         0x85
#define ADDR_CTRL               0x86
#define TIME_ENTER_MONITOR      0x87
#define ADDR_ACTIVE_MODE_RATE   0x88
#define ADDR_MONITOR_MODE_RATE  0x89
#define ADDR_CHIP_CODE          0xA0
#define ADDR_LIBRARY_VERSION    0xA1
#define ADDR_CHIP_ID            0xA3
#define ADDR_G_MODE             0xA4
#define ADDR_POWER_MODE         0xA5
#define ADDR_FIRMARE_ID         0xA6
#define ADDR_FOCALTECH_ID       0xA8
#define ADDR_RELEASE_CODE_ID    0xAF
#define ADDR_STATE              0xBC

#define STAT_TOUCH    0
#define STAT_STREAM   1
#define STAT_RELEASE  2

typedef struct touchPanel_t {
  uint8_t stat ;
  uint16_t x ;
  uint16_t y ;
} tpData_t ;

tpData_t tpData[2] ;
uint8_t readData[8] ;

int btnStat = 0 ;
int touchStat = STAT_RELEASE ;
int touchX = 0 ;
int touchY = 0 ;
int circleColor = RED;

// ▲▼▲▼▲▼ I2C操作用 ▲▼▲▼▲▼
// 1バイト受信
uint8_t ReadByte(uint8_t i2cAddr , uint8_t addr) {
  Wire1.beginTransmission(i2cAddr);
  Wire1.write(addr);
  Wire1.endTransmission();
  Wire1.requestFrom(i2cAddr, 1);
  while (Wire1.available() < 1) ;
  uint8_t val = Wire1.read();
  return val;
}

// 2バイト受信
uint16_t Read2Bytes(uint8_t i2cAddr , uint8_t addr) {
  Wire1.beginTransmission(i2cAddr);
  Wire1.write(addr);
  Wire1.endTransmission();
  Wire1.requestFrom(i2cAddr, 2);
  while (Wire1.available() < 2) ;
  uint8_t valH = Wire1.read();
  uint8_t valL = Wire1.read();
  return valH << 8 | valL ;
}

// 1バイト送信
void WriteByte(uint8_t i2cAddr , uint8_t addr, uint8_t data) {
  Wire1.beginTransmission(i2cAddr);
  Wire1.write(addr);
  Wire1.write(data);
  Wire1.endTransmission();
}

// ▲▼▲▼▲▼ Setup ▲▼▲▼▲▼
void setup() {
  M5.begin();
  M5.Lcd.setBrightness(32);
  M5.Lcd.setRotation(1) ;
  M5.Lcd.fillScreen(WHITE);
  M5.Lcd.setTextFont(4);
  M5.Lcd.setTextColor(BLACK,WHITE);

  Serial.begin(9600) ;
  delay(2000) ;
  Serial.println() ;
  Serial.println() ;

  // --- Power Module : 
  uint8_t val = ReadByte(I2C_ADDR_AXP192,0x12);
  Serial.print("AXP192 Power Register 12h:") ;
  Serial.println(val,HEX) ;
  val |= 0x77;  
  WriteByte(I2C_ADDR_AXP192 , 0x12, val);
  val = ReadByte(I2C_ADDR_AXP192,0x27);
  Serial.print("AXP192 Power Register 27h:") ;
  Serial.println(val,HEX) ;
  val = ReadByte(I2C_ADDR_AXP192,0x28);
  Serial.print("AXP192 Power Register 28h:") ;
  Serial.println(val,HEX) ;
  Serial.println() ;

  // バックライトの明るさ
  val = 0xCC ;
  WriteByte(I2C_ADDR_AXP192 , 0x27, val);

  // Touch Panel
  uint8_t data8 = ReadByte(I2C_ADDR_FT6336,ADDR_DEVICE_MODE) ;
  Serial.print("Device Mode: ");
  switch(data8) {
    case 0x00: Serial.println("WORKING Mode");  break ;
    case 0x04: Serial.println("FACTORY Mode");  break ;
    default: Serial.print("- 0x"); Serial.println(data8,HEX);  
  }

  data8 = ReadByte(I2C_ADDR_FT6336,ADDR_CHIP_CODE) ;
  Serial.print("CHIP Code: ");
  switch(data8) {
    case 0x00: Serial.println("FT6236G");  break ;
    case 0x01: Serial.println("FT6336G");  break ;
    case 0x02: Serial.println("FT6336U");  break ;
    case 0x03: Serial.println("FT6426");  break ;
    default: Serial.print("- 0x"); Serial.println(data8,HEX);  
  }

  data8 = ReadByte(I2C_ADDR_FT6336,ADDR_THRESHOLD) ;
  Serial.print("Threshold(タッチ閾値): 0x");  Serial.println(data8,HEX);  

  data8 = ReadByte(I2C_ADDR_FT6336,ADDR_CTRL) ;
  Serial.print("Control Mode(監視モードへの移行): ");
  switch(data8) {
    case 0x00: Serial.println("禁止");  break ;
    case 0x01: Serial.println("許可");  break ;
    default: Serial.print("- 0x"); Serial.println(data8,HEX);  
  }

  data8 = ReadByte(I2C_ADDR_FT6336,TIME_ENTER_MONITOR) ;
  Serial.print("Time Period for enter to Monitor Mode(モニターモードに切り替わるまでの時間[秒]): ");  Serial.println(data8);  
 
  uint16_t data16 = Read2Bytes(I2C_ADDR_FT6336,ADDR_LIBRARY_VERSION) ;
  Serial.print("LIB Ver(APP ライブラリ ファイルのバージョン): 0x");  Serial.println(data16,HEX);  

  data8 = ReadByte(I2C_ADDR_FT6336,ADDR_G_MODE) ;
  Serial.print("G Mode(レポート ポイントの INT モード制御): ");
  switch(data8) {
    case 0x00: Serial.println("レポート ポイントが延長されない場合の INT ロー時間");  break ;
    case 0x01: Serial.println("レポート ポイントを延長するときの INT ロー タイム");  break ;
    default: Serial.print("- 0x"); Serial.println(data8,HEX);  
  }

  data8 = ReadByte(I2C_ADDR_FT6336,ADDR_POWER_MODE) ;
  Serial.print("POWER Mode: ");
  switch(data8) {
    case 0x00: Serial.println("P_ACTIVE");  break ;
    case 0x01: Serial.println("P_MONITOR");  break ;
    case 0x02: Serial.println("P_STANDBY");  break ;
    case 0x03: Serial.println("P_HIBERNATE");  break ;
    default: Serial.print("- 0x"); Serial.println(data8,HEX);  
  }

  data8 = ReadByte(I2C_ADDR_FT6336,ADDR_FIRMARE_ID) ;
  Serial.print("Firm ID(ファームウエアバージョン): 0x");  Serial.println(data8,HEX);  

  data8 = ReadByte(I2C_ADDR_FT6336,ADDR_RELEASE_CODE_ID) ;
  Serial.print("Release Code ID: 0x");  Serial.println(data8,HEX);  

  data8 = ReadByte(I2C_ADDR_FT6336,ADDR_STATE) ;
  Serial.print("State:");
  switch(data8) {
    case 0x00: Serial.println("InFO mode");  break ;
    case 0x01: Serial.println("Normal working mode");  break ;
    case 0x02: Serial.println("Factory mode");  break ;
    case 0x03: Serial.println("Auto calibration mode");  break ;
    default: Serial.print("- 0x"); Serial.println(data8,HEX);  
  }

  data8 = ReadByte(I2C_ADDR_FT6336,ADDR_ACTIVE_MODE_RATE) ;
  Serial.print("Working mode scan period: 0x");  Serial.println(data8,HEX);  

  data8 = ReadByte(I2C_ADDR_FT6336,ADDR_MONITOR_MODE_RATE) ;
  Serial.print("MONITOR mode scan cycle: 0x");  Serial.println(data8,HEX);  

  // -----
  Serial.println(); 
}

// ▲▼▲▼▲▼ Loop ▲▼▲▼▲▼
void loop() {
  char tempString[128]; 
  uint8_t readPoints = ReadByte(I2C_ADDR_FT6336,ADDR_TD_STATUS) ;
  if (readPoints == 1 || readPoints == 2) {
    if (readPoints == 1) {
        uint16_t data1 = Read2Bytes(I2C_ADDR_FT6336,ADDR_TOUCH_X1) ;
        uint16_t data2 = Read2Bytes(I2C_ADDR_FT6336,ADDR_TOUCH_Y1) ;
        uint8_t id1 = (data2 >> 12) & 0x01 ;
        tpData[id1].stat = (tpData[id1].stat == STAT_RELEASE) ? STAT_TOUCH : STAT_STREAM; 
        tpData[id1].x = data1 & 0x0FFF ;
        tpData[id1].y = data2 & 0x0FFF ;
        tpData[~id1 & 0x01].stat = STAT_RELEASE; 
    } else {
        uint16_t data1 = Read2Bytes(I2C_ADDR_FT6336,ADDR_TOUCH_X1) ;
        uint16_t data2 = Read2Bytes(I2C_ADDR_FT6336,ADDR_TOUCH_Y1) ;
        uint8_t id1 = (data2 >> 12) & 0x01 ;
        tpData[id1].stat = (tpData[id1].stat == STAT_RELEASE) ? STAT_TOUCH : STAT_STREAM; 
        tpData[id1].x = data1 & 0x0FFF ;
        tpData[id1].y = data2 & 0x0FFF ;

        data1 = Read2Bytes(I2C_ADDR_FT6336,ADDR_TOUCH_X2) ;
        data2 = Read2Bytes(I2C_ADDR_FT6336,ADDR_TOUCH_Y2) ;
        uint8_t id2 = (data2 >> 12) & 0x01 ;
        tpData[id2].stat = (tpData[id2].stat == STAT_RELEASE) ? STAT_TOUCH : STAT_STREAM; 
        tpData[id2].x = data1 & 0x0FFF ;
        tpData[id2].y = data2 & 0x0FFF ;
    }
    sprintf(tempString, "Poins %d / TD1(", readPoints); 
    Serial.print(tempString); 
    switch(tpData[0].stat) {
      case 0:  Serial.print("TOUCH  "); break ; 
      case 1:  Serial.print("MOVE   "); break ; 
      default:  Serial.print("RELEASE"); break ; 
    }
    if (tpData[0].stat == STAT_TOUCH || tpData[0].stat == STAT_STREAM) {
      sprintf(tempString, ",%4d,%4d) / TD2(", tpData[0].x, tpData[0].y); 
    } else {
      sprintf(tempString, ",----,----) / TD2("); 
    }
    Serial.print(tempString); 
    switch(tpData[1].stat) {
      case 0:  Serial.print("TOUCH  "); break ; 
      case 1:  Serial.print("MOVE   "); break ; 
      default:  Serial.print("RELEASE"); break ; 
    }
    if (tpData[1].stat == STAT_TOUCH || tpData[1].stat == STAT_STREAM) {
      sprintf(tempString, ",%4d,%4d)", tpData[1].x, tpData[1].y); 
    } else {
      sprintf(tempString, ",----,----)"); 
    }
    Serial.println(tempString); 
  } else {
    if (tpData[0].stat != STAT_RELEASE || tpData[1].stat != STAT_RELEASE) {
      Serial.println("Poins 0 / TD1(RELEASE,----,----) / TD2(RELEASE,----,----)"); 
    }
    tpData[0].stat = STAT_RELEASE; 
    tpData[1].stat = STAT_RELEASE; 
  }

  // ----- タッチパネルとボタンとしての判定
  int btnID = -1 ;
  int touchID = -1 ;
  if (tpData[0].stat != STAT_RELEASE) {
    if (tpData[0].y < 240) {
      touchID = 0 ;
    } else {
      btnID = 0 ;
    }
  }
  if (tpData[1].stat != STAT_RELEASE) {
    if (tpData[1].y < 240) {
      touchID = 1 ;
    } else {
      btnID = 1 ;
    }
  }
  if (btnID != -1) {
    btnStat = 0 ;
    if (30 <= tpData[btnID].x && tpData[btnID].x <= 110) { // 70±40
      btnStat = 1 ; 
    }
    if (130 <= tpData[btnID].x && tpData[btnID].x <= 210) {// 170±40
      btnStat = 2 ; 
    }
    if (230 <= tpData[btnID].x && tpData[btnID].x <= 310) {// 270±40
      btnStat = 3 ;
    }
  } else {
    btnStat = 0 ;  
  }

  switch(btnStat) {
    case 1: circleColor = RED ; break ;
    case 2: circleColor = GREEN ; break ;
    case 3: circleColor = BLUE ; break ;
  }

  if (touchID != -1) {
    M5.Lcd.fillCircle(touchX, touchY, 30, WHITE);
    touchStat = tpData[touchID].stat ;
    touchX =  tpData[touchID].x;
    touchY =  tpData[touchID].y;
    M5.Lcd.fillCircle(touchX, touchY, 30, circleColor);
  } else {
    touchStat = STAT_RELEASE ;
  }

  M5.Lcd.setCursor(10, 10);
  switch(touchStat) {
    case 0:  M5.Lcd.print("TOUCH  "); break ; 
    case 1:  M5.Lcd.print("MOVE   "); break ; 
    default: M5.Lcd.print("RELEASE"); break ; 
  }
  sprintf(tempString, "/POS(%4d,%4d)    ",touchX,touchY); 
  M5.Lcd.print(tempString);
  M5.Lcd.setCursor(10, 40);
  M5.Lcd.print("BTN:");
  switch(btnStat) {
    case 1:  M5.Lcd.print("[A]"); break ; 
    case 2:  M5.Lcd.print("[B]"); break ; 
    case 3:  M5.Lcd.print("[C]"); break ; 
    default: M5.Lcd.print("[-]"); break ; 
  }
  if (btnStat != -1) {
    sprintf(tempString, "/POS(%4d,%4d)    ",tpData[btnID].x,tpData[btnID].y); 
  } else {
    sprintf(tempString, "/POS(----,----)    "); 
  }
  M5.Lcd.print(tempString);
}
